/*==============================================================================
    FILE:           user_usb_lib.c

    DESCRIPTION:

    Copyright (c) 2023 Closed Loop Design, LLC

    This software is supplied "AS IS" without any warranties, express, implied
    or statutory, including but not limited to the implied warranties of fitness
    for purpose, satisfactory quality and non-infringement. Closed Loop Design LLC
    extends you a royalty-free right to use, reproduce and distribute this source
    file as well as executable files created using this source file for use with
    Analog Devices SC5xx family processors only. Nothing else gives you
    the right to use this source file.

==============================================================================*/
/*!
 * @file      user_usb_lib.c
 * @brief
 *
 * @details
 *
 *
 *            Copyright (c) 2023 Closed Loop Design, LLC
 *
 *            This software is supplied "AS IS" without any warranties, express, implied
 *            or statutory, including but not limited to the implied warranties of fitness
 *            for purpose, satisfactory quality and non-infringement. Closed Loop Design LLC
 *            extends you a royalty-free right to use, reproduce and distribute this source
 *            file as well as executable files created using this source file for use with
 *            Analog Devices SC5xx family processors only. Nothing else gives you
 *            the right to use this source file.
 *
 */

#include "user_usb_lib.h"

#include "cld_sc598_audio_2_0_w_cdc_lib.h"
#include "ll_mgr.h"

#include <services/gpio/adi_gpio.h>
#include <services/int/adi_int.h>
#include <services/tmr/adi_tmr.h>
#include <string.h>
#include <stdio.h>

#define SWAP_ENDIAN_4_BYTES(x)   (  (x << 24)                |  \
                                    ((x <<  8) & 0x00ff0000)  | \
                                    ((x >>  8) & 0x0000ff00)  | \
                                     (x >> 24)                  \
                                  )

#define USER_PRODUCT_ID_AUDIO   0x0006  /*!< USB Product ID. */

#define AUDIO_IN_ENABLED        1

#define USER_AUDIO_W_CDC_NUM_CHANNELS   2

/**
  * Feature Unit ID used to control the Headphone volume
  */
#define USER_AUDIO_HEADPHONE_FEATURE_UNIT_ID                    0x09

/**
 * Control and Channel number from the Setup Packet wValue field of
 * the Headphone Master channel
 */
#define USER_AUDIO_HEADPHONE_CONTROL_AND_CHANNEL_NUM_MASTER     0x0200
/**
 * Control and Channel number from the Setup Packet wValue field of
 * the Headphone left channel
 */
#define USER_AUDIO_HEADPHONE_CONTROL_AND_CHANNEL_NUM_LEFT       0x0201
/**
 *Control and Channel number from the Setup Packet wValue field of
 * the Headphone right channel
 */
#define USER_AUDIO_HEADPHONE_CONTROL_AND_CHANNEL_NUM_RIGHT      0x0202

/**
 * Feature Unit ID used to control the Microphone volume
 */
#define USER_AUDIO_MICROPHONE_FEATURE_UNIT_ID                   0x0a
/**
 *Control and Channel number from the Setup Packet wValue field of
 * the Microphone volume
 */
#define USER_AUDIO_MICROPHONE_CONTROL_AND_CHANNEL_NUM_MASTER    0x0200

/**
 * Control and Channel number from the Setup Packet wValue field of
 * the Microphone volume
 */
#define USER_AUDIO_MUTE_CONTROL_MASTER_CHANNEL                  0x0100

/**
 * Clock Source ID connected to the headphones and microphone Units
 */
#define USER_AUDIO_CLOCK_SOURCE_ID_DAC                              0x03
#define USER_AUDIO_CLOCK_SOURCE_ID_ADC                              0x04

/**
 * High Speed max packet size should be large enough to hold 1 bInterval worth
 * of audio samples.
 */
#if (USER_AUDIO_W_CDC_NUM_CHANNELS > 16)
#define USER_AUDIO_MAX_PACKET_SIZE                              1024
#else
#define USER_AUDIO_MAX_PACKET_SIZE                              512
#endif

#define USB_POLLING_INTERVAL_MICROSECONDS       1000
#if (USER_AUDIO_48kHz_SAMPLE_RATE)
#define NUM_AUDIO_SAMPLES                       ((48000/USB_POLLING_INTERVAL_MICROSECONDS))
#else
#define NUM_AUDIO_SAMPLES                       ((44100/USB_POLLING_INTERVAL_MICROSECONDS) + ((44100 % USB_POLLING_INTERVAL_MICROSECONDS)==0 ? 0 : 1))
#endif
#define AUDIO_SAMPLE_SIZE_BYTES                 4


#define AUDIO_IN_BUFFER_SIZE                    (NUM_AUDIO_SAMPLES*AUDIO_SAMPLE_SIZE_BYTES*2)

#pragma pack (1)

/*
   USB Audio v2.0 Unit and Terminal descriptors that describe a simple
   audio device comprised of the following:

    Input Terminal - USB Streaming Endpoint
        ID = 0x01
        Channels: Left, Right
    Input Terminal - Microphone
        ID = 0x02
        Channels: Left, Right
    Output Terminal - Speaker
        ID = 0x06
        Source ID = 0x09
    Output Terminal - USB Streaming Endpoint
        ID = 0x07
        Source ID = 0x0a
    Feature Unit
        ID = 0x09
        Source ID = 0x01
        Controls:
            Master Channel 0: Mute (Control 1)
            Channel 1 (Left): Volume (Control 2)
            Channel 2 (Right): Volume (Control 2)
    Feature Unit
        ID = 0x0a
        Source ID = 0x02
        Controls:
            Master Channel 0: Volume (Control 2)
 */
static const unsigned char user_audio_unit_and_terminal_descriptor[] __attribute__ ((aligned (4), section(".l3_uncached_code")))=  /*!< USB Audio v2.0 Unit and Terminal descriptors that describe a simple audio device. */
{
    /* Input Terminal Descriptor - USB Endpoint */
    0x11,                   /* bLength */
    0x24,                   /* bDescriptorType = Class Specific Interface */
    0x02,                   /* bDescriptorSubType = Input Terminal */
    0x01,                   /* bTerminalID */
    0x01, 0x01,             /* wTerminalType = USB Streaming */
    0x00,                   /* bAssocTerminal */
    0x03,                   /* bCSourceID */
    USER_AUDIO_W_CDC_NUM_CHANNELS,                   /* bNRChannels */

    /* wChannelConfig */
    ((1 << USER_AUDIO_W_CDC_NUM_CHANNELS)-1) & 0xff,
    (((1 << USER_AUDIO_W_CDC_NUM_CHANNELS)-1) >> 8) & 0xff,
    (((1 << USER_AUDIO_W_CDC_NUM_CHANNELS)-1) >> 16) & 0xff,
    (((1 << USER_AUDIO_W_CDC_NUM_CHANNELS)-1) >> 24) & 0xff,

    0x00,                   /* iChannelNames */
    0x00,0x00,              /* bmControls */
    0x00,                   /* iTerminal */
#if (AUDIO_IN_ENABLED == 1)
    /* Input Terminal Descriptor - Microphone */
    0x11,                   /* bLength */
    0x24,                   /* bDescriptorType = Class Specific Interface */
    0x02,                   /* bDescriptorSubType = Input Terminal */
    0x02,                   /* bTerminalID */
    0x01, 0x02,             /* wTerminalType = Microphone */
    0x00,                   /* bAssocTerminal */
    0x04,                   /* bCSourceID */
    USER_AUDIO_W_CDC_NUM_CHANNELS,                   /* bNRChannels */

    /* wChannelConfig */
    ((1 << USER_AUDIO_W_CDC_NUM_CHANNELS)-1) & 0xff,
    (((1 << USER_AUDIO_W_CDC_NUM_CHANNELS)-1) >> 8) & 0xff,
    (((1 << USER_AUDIO_W_CDC_NUM_CHANNELS)-1) >> 16) & 0xff,
    (((1 << USER_AUDIO_W_CDC_NUM_CHANNELS)-1) >> 24) & 0xff,

    0x00,                   /* iChannelNames */
    0x00,0x00,              /* bmControls */
    0x00,                   /* iTerminal */
#endif
    /* Output Terminal Descriptor - Speaker */
    0x0c,                   /* bLength */
    0x24,                   /* bDescriptorType = Class Specific Interface */
    0x03,                   /* bDescriptorSubType = Output Terminal */
    0x06,                   /* bTerminalID */
    0x01, 0x03,             /* wTerminalType - Speaker */
    0x00,                   /* bAssocTerminal */
    0x09,                   /* bSourceID */
    0x03,                   /* bCSourceID */
    0x00, 0x00,             /* bmControls */
    0x00,                   /* iTerminal */
#if (AUDIO_IN_ENABLED == 1)
    /* Output Terminal Descriptor - USB Endpoint */
    0x0c,                   /* bLength */
    0x24,                   /* bDescriptorType = Class Specific Interface */
    0x03,                   /* bDescriptorSubType = Output Terminal */
    0x07,                   /* bTerminalID */
    0x01, 0x01,             /* wTerminalType - USB Streaming */
    0x00,                   /* bAssocTerminal */
    0x0a,                   /* bSourceID */
    0x04,                   /* bCSourceID */
    0x00, 0x00,             /* bmControls */
    0x00,                   /* iTerminal */
#endif
    /* Feature Unit Descriptor */
    //0x2a,                   /* bLength */
    (10 + (4 * USER_AUDIO_W_CDC_NUM_CHANNELS)), /* bLength */
    0x24,                   /* bDescriptorType = Class Specific Interface */
    0x06,                   /* bDescriptorSubType = Feature Unit */
    0x09,                   /* bUnitID */
    0x01,                   /* bSourceID */
    0x00, 0x00, 0x00, 0x00, /* bmaControls - Master */
    0x0f, 0x00, 0x00, 0x00, /* bmaControls - Front Left */
    0x0f, 0x00, 0x00, 0x00, /* bmaControls - Front Right */
    0x00,                   /* iFeature */
    /* Feature Unit Descriptor */
#if (AUDIO_IN_ENABLED == 1)
    (10 + (4 * USER_AUDIO_W_CDC_NUM_CHANNELS)), /* bLength */
    0x24,                   /* bDescriptorType = Class Specific Interface */
    0x06,                   /* bDescriptorSubType = Feature Unit */
    0x0A,                   /* bUnitID */
    0x02,                   /* bSourceID */
    0x00, 0x00, 0x00, 0x00, /* bmaControls - Master */
    0x0d, 0x00, 0x00, 0x00, /* bmaControls - Front Left */
    0x0d, 0x00, 0x00, 0x00, /* bmaControls - Front Right */
    0x00,                   /* iFeature */
#endif
    /* Clock Source Descriptor */
    0x08,                   /* bLength */
    0x24,                   /* bDescriptorType = Class Specific Interface */
    0x0a,                   /* bDescriptorSubType = Clock Source */
    0x03,                   /* ClockID */
    0x01,                   /* bmAttributes - Internal Fixed Clock */
    0x01,                   /* bmControls */
    0x00,                   /* bAssocTerminal */
    0x00,                   /* iClockSource */
    /* Clock Source Descriptor */
    0x08,                   /* bLength */
    0x24,                   /* bDescriptorType = Class Specific Interface */
    0x0a,                   /* bDescriptorSubType = Clock Source */
    0x04,                   /* ClockID */
    0x01,                   /* bmAttributes - Internal Fixed Clock */
    0x01,                   /* bmControls */
    0x00,                   /* bAssocTerminal */
    0x00,                   /* iClockSource */

};

static const unsigned char user_audio_in_stream_format_descriptor[]  __attribute__ ((aligned (4), section(".l3_uncached_code")))=   /*!<  Isochronous IN endpoint PCM format descriptor */
{
    0x06,               /* bLength */
    0x24,               /* bDescriptorType - Class Specific Interface */
    0x02,               /* bDescriptorSubType - Format Type */
    0x01,               /* bFormatType - Format Type 1 */
    0x04,               /* bSubSlotSize */
    24,                 /* bBitResolution */
};

static const unsigned char user_audio_out_stream_format_descriptor[]  __attribute__ ((aligned (4), section(".l3_uncached_code")))=  /*!< Isochronous OUT endpoint PCM format descriptor */
{
    0x06,               /* bLength */
    0x24,               /* bDescriptorType - Class Specific Interface */
    0x02,               /* bDescriptorSubType - Format Type */
    0x01,               /* bFormatType - Format Type 1 */
    0x04,               /* bSubSlotSize */
    24,                 /* bBitResolution */
};


#pragma pack ()

/*!< IN Audio Stream Interface Endpoint Data Descriptor */
static const CLD_Audio_2_0_Audio_Stream_Data_Endpoint_Descriptor user_audio_in_stream_endpoint_desc  __attribute__ ((aligned (4), section(".l3_uncached_code")))=
{
    .b_length = sizeof(CLD_Audio_2_0_Audio_Stream_Data_Endpoint_Descriptor),
    .b_descriptor_type                  = 0x25,             /* Class Specific Endpoint */
    .b_descriptor_subtype               = 0x01,             /* Endpoint - General */
    .bm_attributes                      = 0x00,             /* max packet only set to 0 */
    .bm_controls                        = 0x00,
    .b_lock_delay_units                 = 0x00,             /* Undefined */
    .w_lock_delay                       = 0x00,
};

/*!< OUT Audio Stream Interface Endpoint Data Descriptor */
static const CLD_Audio_2_0_Audio_Stream_Data_Endpoint_Descriptor user_audio_out_stream_endpoint_desc  __attribute__ ((aligned (4), section(".l3_uncached_code")))=
{
    .b_length = sizeof(CLD_Audio_2_0_Audio_Stream_Data_Endpoint_Descriptor),
    .b_descriptor_type                  = 0x25,             /* Class Specific Endpoint */
    .b_descriptor_subtype               = 0x01,             /* Endpoint - General */
    .bm_attributes                      = 0x00,             /* max packet only set to 0 */
    .bm_controls                        = 0x00,
    .b_lock_delay_units                 = 0x02,             /* Milliseconds */
    .w_lock_delay                       = 0x01,             /* 1 Millisecond */
};


/* Function prototypes */
static void user_audio_audio_stream_data_transmitted (void);
static void user_audio_audio_stream_data_transmit_aborted (void);
//static CLD_USB_Transfer_Request_Return_Type user_audio_stream_data_received (CLD_USB_Transfer_Params * p_transfer_data);
static CLD_USB_Data_Received_Return_Type user_audio_stream_data_receive_complete (unsigned int num_bytes);
static CLD_USB_Transfer_Request_Return_Type user_audio_set_req_cmd (CLD_Audio_2_0_Cmd_Req_Parameters * p_req_params, CLD_USB_Transfer_Params * p_transfer_data);
static CLD_USB_Transfer_Request_Return_Type user_audio_get_req_cmd (CLD_Audio_2_0_Cmd_Req_Parameters * p_req_params, CLD_USB_Transfer_Params * p_transfer_data);
static CLD_USB_Data_Received_Return_Type user_audio_set_volume_req (unsigned int num_bytes);
static void user_audio_streaming_rx_endpoint_enabled (CLD_Boolean enabled);
static void user_audio_streaming_tx_endpoint_enabled (CLD_Boolean enabled);

static void user_cdc_usb_in_transfer_complete (void);
static void user_cdc_usb_in_transfer_aborted (void);
static void user_cdc_usb_out_transfer_aborted (void);
static CLD_USB_Data_Received_Return_Type user_cdc_serial_data_rx_transfer_complete (unsigned int num_bytes);
static void user_cdc_serial_data_receive (void);
static CLD_USB_Transfer_Request_Return_Type user_cdc_cmd_send_encapsulated_cmd (CLD_USB_Transfer_Params * p_transfer_data);
static CLD_USB_Transfer_Request_Return_Type user_cdc_cmd_get_encapsulated_resp (CLD_USB_Transfer_Params * p_transfer_data);
static CLD_USB_Data_Received_Return_Type user_cdc_cmd_set_line_coding (CLD_CDC_Line_Coding * p_line_coding);
static CLD_RV user_cdc_cmd_get_line_coding (CLD_CDC_Line_Coding * p_line_coding);
static CLD_USB_Data_Received_Return_Type user_cdc_cmd_set_control_line_state (CLD_CDC_Control_Line_State * p_control_line_state);
static CLD_USB_Data_Received_Return_Type user_cdc_cmd_send_break (unsigned short duration);
static CLD_USB_Data_Transmit_Return_Type user_cdc_tx_serial_data (unsigned long length, unsigned char * p_buffer);

static CLD_RV user_init_usb_phy (void);
static void user_usb_event (CLD_USB_Event event);

static void user_audio_tx_audio_data (void);
static void user_audio_tx_feedback_data (void);
static void user_audio_feedback_xfr_done (void);

static void user_usb_lib_usb_isr(uint32_t Event, void *pArg);

static void user_usb_lib_status (unsigned short status_code, void * p_additional_data, unsigned short additional_data_size);

/*!< Audio Stream IN Interface parameters */
static CLD_Audio_2_0_Stream_Interface_Params user_audio_in_endpoint_params =
{
    .endpoint_number            = 2,                        /* Isochronous endpoint number */
    .max_packet_size_full_speed = USER_AUDIO_MAX_PACKET_SIZE,/* Isochronous endpoint full-speed max packet size */
    .max_packet_size_high_speed = USER_AUDIO_MAX_PACKET_SIZE,/* Isochronous endpoint high-speed max packet size */
    .b_interval_full_speed      = 1,                        /* Isochronous endpoint full-speed bInterval */
    .b_interval_high_speed      = 4,                        /* Isochronous endpoint high-speed bInterval - 1 millisecond */
    .b_terminal_link            = 7,                        /* Terminal ID of the associated Output Terminal */
    .b_format_type              = 1,                        /* Type 1 Format */
    .bm_formats                 = 0x00000001,               /* Type 1 - PCM format */
    .b_nr_channels              = USER_AUDIO_W_CDC_NUM_CHANNELS,/* Number of Channels */
    .bm_channel_config          = ((1 << USER_AUDIO_W_CDC_NUM_CHANNELS)-1), /* Channel Config */
    .p_encoder_descriptor       = CLD_NULL,
    .p_decoder_descriptor       = CLD_NULL,
    .p_format_descriptor        = (unsigned char*)user_audio_in_stream_format_descriptor,
    .p_audio_stream_endpoint_data_descriptor = (CLD_Audio_2_0_Audio_Stream_Data_Endpoint_Descriptor*)&user_audio_in_stream_endpoint_desc,
};

/*!< Audio Stream OUT Interface parameters */
static CLD_Audio_2_0_Stream_Interface_Params user_audio_out_endpoint_params =
{
    .endpoint_number            = 3,                        /* Isochronous endpoint number */
    .max_packet_size_full_speed = USER_AUDIO_MAX_PACKET_SIZE,/* Isochronous endpoint full-speed max packet size */
    .max_packet_size_high_speed = USER_AUDIO_MAX_PACKET_SIZE,/* Isochronous endpoint high-speed max packet size */
    .b_interval_full_speed      = 1,                        /* Isochronous endpoint full-speed bInterval */
    .b_interval_high_speed      = 4,                        /* Isochronous endpoint high-speed bInterval - 1 millisecond */
    .b_terminal_link            = 1,                        /* Terminal ID of the associated Output Terminal */
    .b_format_type              = 1,                        /* Type 1 Format */
    .bm_formats                 = 0x00000001,               /* Type 1 - PCM format */
    .b_nr_channels              = USER_AUDIO_W_CDC_NUM_CHANNELS, /* Number of Channels */
    .bm_channel_config          = ((1 << USER_AUDIO_W_CDC_NUM_CHANNELS)-1), /* Channel Config */
    .p_encoder_descriptor       = CLD_NULL,
    .p_decoder_descriptor       = CLD_NULL,
    .p_format_descriptor        = (unsigned char*)user_audio_out_stream_format_descriptor,
    .p_audio_stream_endpoint_data_descriptor = (CLD_Audio_2_0_Audio_Stream_Data_Endpoint_Descriptor*)&user_audio_out_stream_endpoint_desc,
};

static CLD_Audio_2_0_Rate_Feedback_Params user_audio_rate_feedback_params =
{
    /* Parameters used for the Isochronous Rate Feedback Endpoint Descriptor */
    .max_packet_size_full_speed = 32,
    .max_packet_size_high_speed = 32,
    .b_interval_full_speed      = 1,                        /* Isochronous endpoint full-speed bInterval */
    .b_interval_high_speed      = 4,                        /* Isochronous endpoint high-speed bInterval - 1 millisecond */
};

/*!< CDC Serial Data Bulk OUT endpoint parameters. */
static CLD_Serial_Data_Bulk_Endpoint_Params user_cdc_serial_data_rx_ep_params =
{
    .endpoint_number                = 5,
    .max_packet_size_full_speed     = 64,
    .max_packet_size_high_speed     = 512,
};

/*!< CDC Serial Data Bulk IN endpoint parameters. */
static CLD_Serial_Data_Bulk_Endpoint_Params user_cdc_serial_data_tx_ep_params =
{
    .endpoint_number                = 5,
    .max_packet_size_full_speed     = 64,
    .max_packet_size_high_speed     = 512,
};


/*!< CLD SC59x Audio 2.0 library initialization data. */
static CLD_SC598_Audio_2_0_w_CDC_Lib_Init_Params user_usb_lib_init_params =
{
    .vendor_id                     = 0x064b,       /* Analog Devices Vendor ID */
    .product_id                    = USER_PRODUCT_ID_AUDIO,

    .usb_bus_max_power             = 0,
    .device_descriptor_bcdDevice   = 0x0100,


    .phy_hs_timeout_calibration     = 48/8, /* Round 43 so it is evenly divisable by 8. */
    .phy_fs_timeout_calibration     = 0,
    .phy_delay_req_after_ulip_chirp_cmd = CLD_TRUE,

    .fp_init_usb_phy                = user_init_usb_phy,

    .audio_control_category_code                        = 0x01,                /* Desktop Speaker */

    .p_unit_and_terminal_descriptors                    = (unsigned char*)user_audio_unit_and_terminal_descriptor,
    .unit_and_terminal_descriptors_length               = sizeof(user_audio_unit_and_terminal_descriptor),

    .p_audio_streaming_rx_interface_params              = &user_audio_out_endpoint_params,
    .p_audio_rate_feedback_rx_params                    = &user_audio_rate_feedback_params,

#if (AUDIO_IN_ENABLED == 1)
    .p_audio_streaming_tx_interface_params              = &user_audio_in_endpoint_params,
#else
    .p_audio_streaming_tx_interface_params              = CLD_NULL,
#endif

    .fp_audio_set_req_cmd                               = user_audio_set_req_cmd,
    .fp_audio_get_req_cmd                               = user_audio_get_req_cmd,

    .fp_audio_streaming_rx_endpoint_enabled             = user_audio_streaming_rx_endpoint_enabled,
    .fp_audio_streaming_tx_endpoint_enabled             = user_audio_streaming_tx_endpoint_enabled,

    .p_serial_data_rx_endpoint_params                   = &user_cdc_serial_data_rx_ep_params,
    .p_serial_data_tx_endpoint_params                   = &user_cdc_serial_data_tx_ep_params,

    .fp_cdc_cmd_send_encapsulated_cmd                   = user_cdc_cmd_send_encapsulated_cmd,
    .fp_cdc_cmd_get_encapsulated_resp                   = user_cdc_cmd_get_encapsulated_resp,

    .fp_cdc_cmd_set_line_coding                         = user_cdc_cmd_set_line_coding,
    .fp_cdc_cmd_get_line_coding                         = user_cdc_cmd_get_line_coding,

    .fp_cdc_cmd_set_control_line_state                  = user_cdc_cmd_set_control_line_state,

    .fp_cdc_cmd_send_break                              = user_cdc_cmd_send_break,

    .support_cdc_network_connection                     = 1,
    .cdc_class_bcd_version                              = 0x0120,       /* CDC Version 1.2 */
    .cdc_class_control_protocol_code                    = 0,            /* No Class Specific protocol */

    .p_usb_string_manufacturer                          = "Analog Devices Inc",
    .p_usb_string_product                               = "SC598 USB Audio 2.0 w CDC Device",
    .p_usb_string_serial_number                         = CLD_NULL,
    .p_usb_string_configuration                         = CLD_NULL,
    .p_usb_string_audio_control_interface               = CLD_NULL,
    .p_usb_string_audio_streaming_out_interface         = "Audio 2.0 Output",
    .p_usb_string_audio_streaming_in_interface          = "Audio 2.0 Input",
    .p_usb_string_communication_class_interface         = "CLD CDC Ctrl",
    .p_usb_string_data_class_interface                  = "CLD CDC Data",

    .user_string_descriptor_table_num_entries           = 0,
    .p_user_string_descriptor_table                     = NULL,

    .usb_string_language_id                             = 0x0409,       /* English (US) language ID */

    .fp_cld_usb_event_callback                          = user_usb_event,

    .fp_cld_lib_status                                  = user_usb_lib_status,
};

#pragma pack(1)
#define USB_AUDIO_MAX_NUM_SUB_RANGES     1  /*!< Maximum number of sub-range values the device supports */

/**
 * Structure used to report 32-bit USB Audio 2.0 range values.
 */
typedef struct
{
    unsigned short  w_num_sub_ranges;   /*!< The number of sub ranges being reported */
    struct
    {
        unsigned int   d_min;           /*!< Range Minimum */
        unsigned int   d_max;           /*!< Range Minimum */
        unsigned int   d_res;           /*!< Range Minimum */
    } sub_ranges[USB_AUDIO_MAX_NUM_SUB_RANGES]; /*!< Array of supported sub-ranges */
} USB_Audio_4_byte_Range_Response;

/**
 * Structure used to report 16-bit USB Audio 2.0 range values.
 */
typedef struct
{
    unsigned short  w_num_sub_ranges;   /*!< The number of sub-ranges being reported */
    struct
    {
        unsigned short  d_min;          /*!< Range Minimum */
        unsigned short  d_max;          /*!< Range Minimum */
        unsigned short  d_res;          /*!< Range Minimum */
    } sub_ranges[USB_AUDIO_MAX_NUM_SUB_RANGES]; /*!< Array of supported sub-ranges */
} USB_Audio_2_byte_Range_Response;

static USB_Audio_2_byte_Range_Response user_audio_2_byte_range_resp __attribute__ ((aligned (4), section(".l3_uncached_bss")));

/**
 * Structure used to report 8-bit USB Audio 2.0 range values.
 */
typedef struct
{
    unsigned short  w_num_sub_ranges;   /*!< The number of sub-ranges being reported */
    struct
    {
        unsigned char  d_min;           /*!< Range Minimum */
        unsigned char  d_max;           /*!< Range Minimum */
        unsigned char  d_res;           /*!< Range Minimum */
    } sub_ranges[USB_AUDIO_MAX_NUM_SUB_RANGES]; /*!< Array of supported sub-ranges */
} USB_Audio_1_byte_Range_Response;
#pragma pack()

/**
 * Structure used to store the current USB volume settings.
 */
typedef struct
{
    unsigned short master __attribute__ ((aligned (4)));              /*!< Master Volume setting */
    unsigned int vol[USER_AUDIO_W_CDC_NUM_CHANNELS];
    unsigned short min __attribute__ ((aligned (4)));                 /*!< Volume range minimum setting */
    unsigned short max __attribute__ ((aligned (4)));                 /*!< Volume range maximum setting */
    unsigned short resolution __attribute__ ((aligned (4)));          /*!< Volume range resolution */
    unsigned char mute __attribute__ ((aligned (4)));                 /*!< Mute setting */
} User_Audio_Data_Volume;

/**
 * Structure used to store the requested USB volume settings.
 */
typedef struct
{
    unsigned short vol[USER_AUDIO_W_CDC_NUM_CHANNELS] __attribute__ ((aligned (4)));
    unsigned char mute __attribute__ ((aligned (4)));                 /*!< Mute setting */
} User_Audio_Req_Volume;

/**
 * User Audio data
 */
typedef struct
{
    CLD_Boolean             volume_changed;                 /*!< CLD_TRUE when a Set Volume or
                                                                 Mute command is received */
    User_Audio_Data_Volume  headphone_output_volume;        /*!< Current Volume settings */
    User_Audio_Req_Volume   headphone_req_output_volume;    /*!< Requested Volume settings */
    User_Audio_Data_Volume  mic_input_volume;               /*!< Current Microphone Volume Settings */
    User_Audio_Req_Volume   microphone_req_input_volume;    /*!< Requested Microphone Volume */
    CLD_Boolean isochronous_in_enabled;                     /*!< CLD_TRUE - Isochronous IN endpoint enabled */
    CLD_Boolean isochronous_in_idle;                        /*!< CLD_TRUE - Endpoint is idle so data can
                                                                            be transmitted. */
    unsigned int clock_sample_rate;                        /*!< Active audio sample rate */
    CLD_Boolean isochronous_out_enabled;                    /*!< CLD_TRUE - Isochronous OUT endpoint enabled */
    CLD_Boolean rate_feedback_idle;
} User_Audio_Data;

/* Locate the User_Audio_Data in un-cached memory since USB DMA will be used to access the
   audio data buffers. */
static User_Audio_Data user_audio_data __attribute__ ((aligned (4), section(".l3_uncached_data"))) =
{
    .volume_changed = CLD_FALSE,
    .headphone_output_volume.min = 0xa060,                  /* ADAU1962A USB Minimum Volume Setting = -95.6250dB */
    .headphone_output_volume.max = 0x0000,                  /* ADAU1962A USB Maximum Volume Setting = 0dB */
    .headphone_output_volume.resolution = 0x0060,           /* ADAU1962A USB Volume Resolution Setting = 0.3750dB */

    .mic_input_volume.min = 0xDC60,                         /* ADAU1979 USB Minimum Volume Setting = -35.625dB */
    .mic_input_volume.max = 0x3C00,                         /* ADAU1979 USB Maximum Volume Setting = 60dB */
    .mic_input_volume.resolution = 0x0060,                  /* ADAU1979 USB Volume Resolution Setting = 0.3750dB */

    .isochronous_in_enabled = CLD_FALSE,
    .isochronous_in_idle = CLD_TRUE,
#if (USER_AUDIO_48kHz_SAMPLE_RATE)
    .clock_sample_rate = 48000,
#else
    .clock_sample_rate = 44100,
#endif
    .isochronous_out_enabled = CLD_FALSE,
    .rate_feedback_idle = CLD_TRUE,
};

/* Clock Source sample rate range. */
static USB_Audio_4_byte_Range_Response user_audio_get_clock_sample_rate_range_resp __attribute__ ((aligned (4), section(".l3_uncached_data")))=
{
    .w_num_sub_ranges = 1,
#if (USER_AUDIO_48kHz_SAMPLE_RATE)
    .sub_ranges[0].d_min = 48000,
    .sub_ranges[0].d_max = 48000,
#else
    .sub_ranges[0].d_min = 44100,
    .sub_ranges[0].d_max = 44100,
#endif
    .sub_ranges[0].d_res = 0,
};

/* Current CDC line coding settings (default to 115.2k 8 bits, 1 Stop bit, no parity) */
static CLD_CDC_Line_Coding user_cdc_line_coding __attribute__ ((aligned (4), section(".l3_uncached_data")))=
{
    .data_terminal_rate = 115200,
    .num_stop_bits      = CLD_CDC_LINE_CODING_1_STOP_BITS,
    .parity             = CLD_CDC_LINE_CODING_PARITY_NONE,
    .num_data_bits      = 8,
};


#define USER_CDC_SERIAL_BUFFER_SIZE     512                 /*!< CDC Serial Data receive buffer size */
/**
 * Buffer used to store the received serial data.
 */
typedef struct
{
    unsigned char resp_hdr[12]; /* Must be a multiple of 4-bytes to maintain USB library required 4-byte alignment */
    unsigned char data[USER_CDC_SERIAL_BUFFER_SIZE];
} CDC_Serial_Data;
static CDC_Serial_Data user_cdc_serial_data_rx_buffer __attribute__ ((aligned (4), section(".l3_uncached_data"))) =
{
    .resp_hdr = "\n\rLib Echo: ",
};

static CLD_Boolean first_feedback = CLD_TRUE;

static LL_Mgr_Data user_audio_out_ll;
static LL_Mgr_Data user_audio_in_ll;

static LL_Mgr_Node * p_user_audio_out_ll_node = 0;
static LL_Mgr_Node * p_user_audio_in_ll_node = 0;
static LL_Mgr_Node * p_user_audio_zero_ll_node;

static unsigned char user_audio_num_in_packets = 0;

/**
 * Initializes the CLD USB library.
 *
 * @retval USER_AUDIO_INIT_SUCCESS.
 * @retval USER_AUDIO_INIT_ONGOING.
 * @retval USER_AUDIO_INIT_FAILED.
 */
User_Init_Return_Code user_usb_lib_init (void)
{
    static unsigned char state = 0;
    User_Init_Return_Code rv = USER_INIT_ONGOING;
    CLD_RV cld_rv;
    unsigned int i;

    switch (state)
    {
        case 0:
            /* Configure the GPIO pins connected to the SC589 EZ-Board general purpose LEDs (LED10, LED11, LED12) */
            adi_gpio_SetDirection(ADI_GPIO_PORT_C, ADI_GPIO_PIN_3, ADI_GPIO_DIRECTION_OUTPUT);
            adi_gpio_Clear(ADI_GPIO_PORT_C, ADI_GPIO_PIN_3);
            adi_gpio_SetDirection(ADI_GPIO_PORT_C, ADI_GPIO_PIN_2, ADI_GPIO_DIRECTION_OUTPUT);
            adi_gpio_Clear(ADI_GPIO_PORT_C, ADI_GPIO_PIN_2);
            adi_gpio_SetDirection(ADI_GPIO_PORT_C, ADI_GPIO_PIN_1, ADI_GPIO_DIRECTION_OUTPUT);
            adi_gpio_Clear(ADI_GPIO_PORT_C, ADI_GPIO_PIN_1);

            state++;
        break;
        case 1:
            /* Initialize the CLD Audio v2.0 with CDC Library */
            cld_rv = cld_sc598_audio_2_0_w_cdc_lib_init(&user_usb_lib_init_params);

            if (cld_rv == CLD_SUCCESS)
            {

                adi_int_InstallHandler(INTR_USBC0_INT, user_usb_lib_usb_isr, NULL, 1);

                ll_mgr_init(&user_audio_out_ll);
                ll_mgr_init(&user_audio_in_ll);

                /* Get the current volume settings for Headphone Left and Right channels */
                for (i = 0; i < USER_AUDIO_W_CDC_NUM_CHANNELS; i++)
                {
                    /* TODO: Initialize using CODEC volume */

                    user_audio_data.headphone_req_output_volume.vol[i] = user_audio_data.headphone_output_volume.vol[i];
                }

                /* Get the current Mute settings for Headphones */
                /* TODO: Initialize using CODEC setting */
                user_audio_data.headphone_output_volume.mute = 0;

                for (i = 0; i < 2; i++)
                {
                    /* TODO: Initialize using CODEC volume */

                    /* Initialize the requested volume values with the current volume settings */
                    user_audio_data.microphone_req_input_volume.vol[i] = user_audio_data.mic_input_volume.vol[i];
                }
                for (i = 4; i < USER_AUDIO_W_CDC_NUM_CHANNELS; i++)
                {
                    user_audio_data.microphone_req_input_volume.vol[i] = user_audio_data.mic_input_volume.vol[0];
                }

                /* Connect to the USB Host */
                cld_lib_usb_connect();


                rv = USER_INIT_SUCCESS;
            }
            else if (cld_rv == CLD_FAIL)
            {
                rv = USER_INIT_FAILED;
            }
            else
            {
                rv = USER_INIT_ONGOING;
            }
        break;
    }
    return rv;
}


/**
 * Example User Mainline function.
 *
 */
void user_usb_lib_main (void)
{
    static CLD_Time main_time = 0;
    unsigned int i;

    if (cld_time_passed_ms(main_time) >= 250u)
    {
        main_time = cld_time_get();
        adi_gpio_Toggle(ADI_GPIO_PORT_C, ADI_GPIO_PIN_3);
    }
    cld_sc598_audio_2_0_w_cdc_lib_main();

    /* If a Set Volume or Set Mute command data has been received */
    if (user_audio_data.volume_changed == CLD_TRUE)
    {
        user_audio_data.volume_changed = CLD_FALSE;
        for (i = 0; i < USER_AUDIO_W_CDC_NUM_CHANNELS; i++)
        {
            if (user_audio_data.headphone_req_output_volume.vol[i] != user_audio_data.headphone_output_volume.vol[i])
            {
                user_audio_data.headphone_output_volume.vol[i] = user_audio_data.headphone_req_output_volume.vol[i];

                /* TODO: Set CODEC volume */
            }
        }

        for (i = 0; i < USER_AUDIO_W_CDC_NUM_CHANNELS; i++)
        {
            if (user_audio_data.microphone_req_input_volume.vol[i] != user_audio_data.mic_input_volume.vol[i])
            {
                user_audio_data.mic_input_volume.vol[i] = user_audio_data.microphone_req_input_volume.vol[i];
                /* TODO: Set CODEC volume */
            }
        }
    }

    user_audio_tx_feedback_data();

    user_audio_tx_audio_data();
}

/**
 * Transmits the pending audio IN data to the USB Host.
 *
 */
static void user_audio_tx_audio_data (void)
{
    static CLD_USB_Transfer_Params audio_data_tx_params;

#if (AUDIO_IN_ENABLED == 1)
   // adi_rtl_disable_interrupts();

    /* If the Isochronous IN endpoint is enabled and idle */
    if ((user_audio_data.isochronous_in_enabled == CLD_TRUE) &&
        (user_audio_data.isochronous_in_idle == CLD_TRUE))
    {
        /* If there is headphone data waiting to be sent */
        if (user_audio_in_ll.num_elements)
        {
            p_user_audio_in_ll_node = ll_mgr_remove(&user_audio_in_ll);

            /* Transmit the Headphone data using the Audio Streaming IN interface from the tail of the microphone
               circular queue */
            audio_data_tx_params.num_bytes = p_user_audio_in_ll_node->size;
            audio_data_tx_params.p_data_buffer = p_user_audio_in_ll_node->data;
            audio_data_tx_params.callback.fp_usb_in_transfer_complete = user_audio_audio_stream_data_transmitted;
            audio_data_tx_params.transfer_timeout_ms = 100;
            audio_data_tx_params.fp_transfer_aborted_callback = user_audio_audio_stream_data_transmit_aborted;
            if (cld_audio_2_0_lib_transmit_audio_data (&audio_data_tx_params) == CLD_USB_TRANSMIT_SUCCESSFUL)
            {
                user_audio_data.isochronous_in_idle = CLD_FALSE;
            }
        }
    }
    //adi_rtl_reenable_interrupts();
#endif
}

/**
 * Transmits the current feedback value to the USB Host.
 *
 */
static void user_audio_tx_feedback_data (void)
{
    static CLD_USB_Audio_Feedback_Params feedback_transfer_data =
    {
        .fp_transfer_aborted_callback = user_audio_feedback_xfr_done,
        .transfer_timeout_ms = 1100,
        .fp_usb_in_transfer_complete = user_audio_feedback_xfr_done,
    };

    //adi_rtl_disable_interrupts();
    if (user_audio_data.isochronous_out_enabled == CLD_TRUE &&
            user_audio_data.rate_feedback_idle == CLD_TRUE)
    {
        if (first_feedback == CLD_FALSE)
        {
            /* TODO: calculate feedback based on CODEC buffer */

            #if (USER_AUDIO_48kHz_SAMPLE_RATE)
            feedback_transfer_data.desired_data_rate = 48;
            #else
            feedback_transfer_data.desired_data_rate = 44.1;
            #endif
        }
        else
        {
            #if (USER_AUDIO_48kHz_SAMPLE_RATE)
            feedback_transfer_data.desired_data_rate = 48;
            #else
            feedback_transfer_data.desired_data_rate = 44.1;
            #endif
        }

        if (cld_audio_2_0_lib_transmit_audio_rate_feedback_data(&feedback_transfer_data)== CLD_USB_TRANSMIT_SUCCESSFUL)
        {
            user_audio_data.rate_feedback_idle = CLD_FALSE;
        }
    }
    //adi_rtl_reenable_interrupts();
}

/**
 * Function called when a microphone data packet has been transmitted
 * to the USB Host.
 *
 */
static void user_audio_audio_stream_data_transmitted (void)
{
    unsigned int i;

    ll_mgr_free_node(p_user_audio_in_ll_node);
    p_user_audio_in_ll_node = 0;

    user_audio_num_in_packets = 0xff; /* Flag that a Audio IN packet has been sent to the Host */

    user_audio_data.isochronous_in_idle = CLD_TRUE;

    if ((user_audio_data.isochronous_in_enabled == CLD_TRUE))
    {
        /* If there isn't headphone data waiting to be sent sent a zero packet */
        if (user_audio_in_ll.num_elements == 0)
        {
            p_user_audio_zero_ll_node = ll_mgr_get_node(0);

            if (p_user_audio_zero_ll_node)
            {
                p_user_audio_zero_ll_node->size = AUDIO_IN_BUFFER_SIZE;

                for (i = 0; i < AUDIO_IN_BUFFER_SIZE; i++)
                {
                    p_user_audio_zero_ll_node->data[i] = 0;
                }

                user_audio_transmit_audio_stream_data(p_user_audio_zero_ll_node);
            }
        }
    }

    user_audio_tx_audio_data();
}

/**
 * Function called when a the microphone data transmission fails..
 *
 */
static void user_audio_audio_stream_data_transmit_aborted (void)
{
    ll_mgr_free_node(p_user_audio_in_ll_node);
    p_user_audio_in_ll_node = 0;
    user_audio_data.isochronous_in_idle = CLD_TRUE;
    user_audio_tx_audio_data();
}

/**
 * This function primes the Isochronous OUT endpoint so it will
 * receive data.
 */
static void user_audio_stream_data_receive (void)
{
    static CLD_USB_Transfer_Params transfer_params;
    CLD_USB_Data_Receive_Return_Type rv;

    p_user_audio_out_ll_node = ll_mgr_get_node(0);

    if (p_user_audio_out_ll_node)
    {
        transfer_params.num_bytes = USER_AUDIO_MAX_PACKET_SIZE;
        transfer_params.p_data_buffer = p_user_audio_out_ll_node->data;
        transfer_params.transfer_timeout_ms = 0;
        transfer_params.fp_transfer_aborted_callback = CLD_NULL;
        transfer_params.callback.fp_usb_out_transfer_complete = user_audio_stream_data_receive_complete;

        rv = cld_audio_2_0_lib_receive_stream_data(&transfer_params);
        if (rv == CLD_USB_RECEIVE_FAILED_MISALIGNED)
        {
            /* CLD library requires transfer_params.p_data_buffer to be 32-bit aligned */
        }
        else if (rv == CLD_USB_RECEIVE_FAILED)
        {
            /* Isochronous OUT endpoint busy */
        }
    }
}

/**
 * Function called when headphone data has been received.
 *
 * @retval CLD_USB_DATA_GOOD
 * @retval CLD_USB_DATA_BAD_STALL
 */
static CLD_USB_Data_Received_Return_Type user_audio_stream_data_receive_complete (unsigned int num_bytes)
{
    user_audio_stream_data_receive(); /* Prime the OUT endpoint to receive the next steaming audio data packet. */

    p_user_audio_out_ll_node->size = num_bytes;
    user_audio_transmit_audio_stream_data(p_user_audio_out_ll_node);
    p_user_audio_out_ll_node = 0;
    return CLD_USB_DATA_GOOD;
}

/**
 * This function is called by the CLD Audio v2.0 library when a Set Command
 * request is received. This function sets the p_transfer_data parameters to
 * select where the received data should be stored, and what function should
 * be called when the transfer is complete.
 *
 * @param p_transfer_data                - Pointer to the Control OUT transfer data.
 * @param p_transfer_data->num_bytes     - Number of Control OUT bytes.
 * @param p_transfer_data->p_data_buffer - Set to the address where to
 *                                         store the Control OUT data.
 * @param p_transfer_data->callback.usb_in_transfer_complete -
 *                                         Function called when the
 *                                         Control OUT bytes have been
 *                                         received.
 * @param p_transfer_data->transfer_aborted_callback -
 *                                         Optional function that is
 *                                         called if the transfer is
 *                                         aborted.
 * @param p_req_params->req              - Setup Packet bRequest value.
 * @param p_req_params->entity_id        - Requested entity ID (Unit ID,
 *                                         Terminal ID, etc)
 * @param p_req_params->interface_or_endpoint_num -
 *                                         Requested interface or endpoint
 *                                         number depending on the entity.
 * @param p_req_params->setup_packet_wValue
 *
 * @retval CLD_USB_TRANSFER_ACCEPT - Store the data using the p_transfer_data
 *                                   parameters.
 * @retval CLD_USB_TRANSFER_PAUSE - The device isn't ready to process this
 *                                  out packet so pause the transfer
 *                                  until the cld_sc57x_audio_2_0_w_cdc_lib_resume_paused_control_transfer
 *                                  function is called.
 * @retval CLD_USB_TRANSFER_DISCARD - Discard this packet.
 * @retval CLD_USB_TRANSFER_STALL - Stall the OUT endpoint.
 *
 */
static CLD_USB_Transfer_Request_Return_Type user_audio_set_req_cmd (CLD_Audio_2_0_Cmd_Req_Parameters * p_req_params, CLD_USB_Transfer_Params * p_transfer_data)
{
    CLD_USB_Transfer_Request_Return_Type rv = CLD_USB_TRANSFER_DISCARD;

    /* If the USB Host is changing the Headphone Volume */
    if (p_req_params->entity_id == USER_AUDIO_HEADPHONE_FEATURE_UNIT_ID)
    {
        if (p_req_params->req == CLD_REQ_CURRENT)
        {
            if (p_req_params->setup_packet_wValue == USER_AUDIO_HEADPHONE_CONTROL_AND_CHANNEL_NUM_MASTER) /* Master Volume */
            {
                /* Store the volume setting in the headphone_req_output_volume structure */
                p_transfer_data->p_data_buffer = (unsigned char *)&user_audio_data.headphone_req_output_volume.vol[0];
                p_transfer_data->num_bytes = 2;
            }
            else if (((p_req_params->setup_packet_wValue & 0xff00) == 0x0200) &&
                     ((p_req_params->setup_packet_wValue & 0x00ff) <= USER_AUDIO_W_CDC_NUM_CHANNELS))
            {
                /* Store the volume setting in the headphone_req_output_volume structure */
                p_transfer_data->p_data_buffer = (unsigned char *)&user_audio_data.headphone_req_output_volume.vol[(p_req_params->setup_packet_wValue & 0x00ff)-1];
                p_transfer_data->num_bytes = 2;
            }
            else /* Mute */
            {
                /* Store the mute setting in the headphone_req_output_volume structure */
                p_transfer_data->p_data_buffer = (unsigned char *)&user_audio_data.headphone_req_output_volume.mute;
                p_transfer_data->num_bytes = 1;
            }
            p_transfer_data->transfer_timeout_ms = 0;
            p_transfer_data->fp_transfer_aborted_callback = CLD_NULL;
            p_transfer_data->callback.fp_usb_out_transfer_complete = user_audio_set_volume_req;
            rv = CLD_USB_TRANSFER_ACCEPT;
        }
    }
    /* If the USB Host is changing the Microphone Volume */
    else if (p_req_params->entity_id == USER_AUDIO_MICROPHONE_FEATURE_UNIT_ID)
    {
        if (p_req_params->req == CLD_REQ_CURRENT)
        {
            if (p_req_params->setup_packet_wValue == USER_AUDIO_HEADPHONE_CONTROL_AND_CHANNEL_NUM_MASTER) /* Master Volume */
            {
                /* Store the volume setting in the headphone_req_output_volume structure */
                p_transfer_data->p_data_buffer = (unsigned char *)&user_audio_data.microphone_req_input_volume.vol[0];
                p_transfer_data->num_bytes = 2;
            }
            else if (((p_req_params->setup_packet_wValue & 0xff00) == 0x0200) &&
                     ((p_req_params->setup_packet_wValue & 0x00ff) <= USER_AUDIO_W_CDC_NUM_CHANNELS))
            {
                /* Store the volume setting in the headphone_req_output_volume structure */
                p_transfer_data->p_data_buffer = (unsigned char *)&user_audio_data.microphone_req_input_volume.vol[(p_req_params->setup_packet_wValue & 0x00ff)-1];
                p_transfer_data->num_bytes = 2;
            }
            else /* Mute */
            {
                /* Store the mute setting in the headphone_req_output_volume structure */
                p_transfer_data->p_data_buffer = (unsigned char *)&user_audio_data.microphone_req_input_volume.mute;
                p_transfer_data->num_bytes = 1;
            }
            p_transfer_data->transfer_timeout_ms = 0;
            p_transfer_data->fp_transfer_aborted_callback = CLD_NULL;
            p_transfer_data->callback.fp_usb_out_transfer_complete = user_audio_set_volume_req;
            rv = CLD_USB_TRANSFER_ACCEPT;
        }
    }
    return rv;
}

/**
 * This function is called by the CLD Audio library when a Set Volume request data
 * has been received.
 *
 * @retval CLD_USB_DATA_GOOD - Received data is valid.
 * @retval CLD_USB_DATA_BAD_STALL - Received data is invalid.
 */
static CLD_USB_Data_Received_Return_Type user_audio_set_volume_req (unsigned int num_bytes)
{
    /* Flag that new volume settings have been received. */
    user_audio_data.volume_changed = CLD_TRUE;
    return CLD_USB_DATA_GOOD;
}

/**
 * This function is called by the CLD Audio v2.0 library when a Get Command
 * request is received. This function sets the p_transfer_data parameters to
 * select where the transmit data should be sourced, and what function should
 * be called when the transfer is complete.
 *
 * @param p_transfer_data                - Pointer to the Control IN transfer data.
 * @param p_transfer_data->num_bytes     - Number of Control IN bytes.
 * @param p_transfer_data->p_data_buffer - Set to the address where to
 *                                         source the Control IN data.
 * @param p_transfer_data->callback.usb_in_transfer_complete -
 *                                         Function called when the
 *                                         Control IN bytes have been
 *                                         sent.
 * @param p_transfer_data->transfer_aborted_callback -
 *                                         Optional function that is
 *                                         called if the transfer is
 *                                         aborted.
 * @param p_req_params->req              - Setup Packet bRequest value.
 * @param p_req_params->entity_id        - Requested entity ID (Unit ID,
 *                                         Terminal ID, etc)
 * @param p_req_params->interface_or_endpoint_num -
 *                                         Requested interface or endpoint
 *                                         number depending on the entity.
 * @param p_req_params->setup_packet_wValue
 *
 * @retval LD_USB_TRANSFER_ACCEPT - Store the data using the p_transfer_data
 *                                  parameters.
 * @retval CLD_USB_TRANSFER_PAUSE - The device isn't ready to process this
 *                                  out packet so pause the transfer
 *                                  until the cld_sc57x_audio_2_0_w_cdc_lib_resume_paused_control_transfer
 *                                  function is called.
 * @retval CLD_USB_TRANSFER_DISCARD - Discard this packet.
 * @retval CLD_USB_TRANSFER_STALL - Stall the OUT endpoint.
 *
 */
static CLD_USB_Transfer_Request_Return_Type user_audio_get_req_cmd (CLD_Audio_2_0_Cmd_Req_Parameters * p_req_params, CLD_USB_Transfer_Params * p_transfer_data)
{
    CLD_USB_Transfer_Request_Return_Type rv = CLD_USB_TRANSFER_DISCARD;

    /* If the USB Host is Requesting the Headphone Volume */
    if (p_req_params->entity_id == USER_AUDIO_HEADPHONE_FEATURE_UNIT_ID)
    {
        /* Current Setting */
        if (p_req_params->req == CLD_REQ_CURRENT)
        {
            if (p_req_params->setup_packet_wValue == USER_AUDIO_HEADPHONE_CONTROL_AND_CHANNEL_NUM_MASTER) /* Master Volume */
            {
                p_transfer_data->p_data_buffer = (unsigned char *)&user_audio_data.headphone_output_volume.vol[0];
                p_transfer_data->num_bytes = 2;
            }
            else if (((p_req_params->setup_packet_wValue & 0xff00) == 0x0200) &&
                     ((p_req_params->setup_packet_wValue & 0x00ff) <= USER_AUDIO_W_CDC_NUM_CHANNELS))
            {
                p_transfer_data->p_data_buffer = (unsigned char *)&user_audio_data.headphone_output_volume.vol[(p_req_params->setup_packet_wValue & 0x00ff)-1];
                p_transfer_data->num_bytes = 2;
            }
            else /* Mute */
            {
                p_transfer_data->p_data_buffer = (unsigned char *)&user_audio_data.headphone_output_volume.mute;
                p_transfer_data->num_bytes = 1;
            }

            p_transfer_data->transfer_timeout_ms = 0;
            p_transfer_data->fp_transfer_aborted_callback = CLD_NULL;
            p_transfer_data->callback.fp_usb_in_transfer_complete = CLD_NULL;
            rv = CLD_USB_TRANSFER_ACCEPT;
        }
        /* Headphone Range Settings */
        else if (p_req_params->req == CLD_REQ_RANGE)
        {
            user_audio_2_byte_range_resp.w_num_sub_ranges = 1;
            user_audio_2_byte_range_resp.sub_ranges[0].d_max = user_audio_data.headphone_output_volume.max;
            user_audio_2_byte_range_resp.sub_ranges[0].d_min = user_audio_data.headphone_output_volume.min;
            user_audio_2_byte_range_resp.sub_ranges[0].d_res = user_audio_data.headphone_output_volume.resolution;

            p_transfer_data->p_data_buffer = (unsigned char *)&user_audio_2_byte_range_resp;
            p_transfer_data->num_bytes = sizeof(user_audio_2_byte_range_resp);
            p_transfer_data->transfer_timeout_ms = 0;
            p_transfer_data->fp_transfer_aborted_callback = CLD_NULL;
            p_transfer_data->callback.fp_usb_in_transfer_complete = CLD_NULL;
            rv = CLD_USB_TRANSFER_ACCEPT;
        }
    }
    /* If the USB Host is Requesting the Microphone Volume */
    else if (p_req_params->entity_id == USER_AUDIO_MICROPHONE_FEATURE_UNIT_ID)
    {
        /* Current microphone Volume */
        if (p_req_params->req == CLD_REQ_CURRENT)
        {
            if (p_req_params->setup_packet_wValue == USER_AUDIO_MICROPHONE_CONTROL_AND_CHANNEL_NUM_MASTER) /* Master Volume */
            {
                p_transfer_data->p_data_buffer = (unsigned char *)&user_audio_data.mic_input_volume.vol[0];
                p_transfer_data->num_bytes = 2;
            }
            else if (((p_req_params->setup_packet_wValue & 0xff00) == 0x0200) &&
                     ((p_req_params->setup_packet_wValue & 0x00ff) <= USER_AUDIO_W_CDC_NUM_CHANNELS))
            {
                p_transfer_data->p_data_buffer = (unsigned char *)&user_audio_data.mic_input_volume.vol[(p_req_params->setup_packet_wValue & 0x00ff)-1];
                p_transfer_data->num_bytes = 2;
            }
            else /* Mute */
            {
                p_transfer_data->p_data_buffer = (unsigned char *)&user_audio_data.mic_input_volume.mute;
                p_transfer_data->num_bytes = 1;
            }
            p_transfer_data->transfer_timeout_ms = 0;
            p_transfer_data->fp_transfer_aborted_callback = CLD_NULL;
            p_transfer_data->callback.fp_usb_in_transfer_complete = CLD_NULL;
            rv = CLD_USB_TRANSFER_ACCEPT;

        }
        /* Minimum microphone setting */
        else if (p_req_params->req == CLD_REQ_RANGE)
        {
            user_audio_2_byte_range_resp.w_num_sub_ranges = 1;
            user_audio_2_byte_range_resp.sub_ranges[0].d_max = user_audio_data.mic_input_volume.max;
            user_audio_2_byte_range_resp.sub_ranges[0].d_min = user_audio_data.mic_input_volume.min;
            user_audio_2_byte_range_resp.sub_ranges[0].d_res = user_audio_data.mic_input_volume.resolution;

            p_transfer_data->p_data_buffer = (unsigned char *)&user_audio_2_byte_range_resp;
            p_transfer_data->num_bytes = sizeof(user_audio_2_byte_range_resp);
            p_transfer_data->transfer_timeout_ms = 0;
            p_transfer_data->fp_transfer_aborted_callback = CLD_NULL;
            p_transfer_data->callback.fp_usb_in_transfer_complete = CLD_NULL;
            rv = CLD_USB_TRANSFER_ACCEPT;
        }
    }
    /* If the USB Host is Requesting the Clock Source data */
    else if (p_req_params->entity_id == USER_AUDIO_CLOCK_SOURCE_ID_DAC)
    {
        if (p_req_params->req == CLD_REQ_CURRENT)
        {
            p_transfer_data->p_data_buffer = (unsigned char *)&user_audio_data.clock_sample_rate;
            p_transfer_data->num_bytes = 4;
            p_transfer_data->transfer_timeout_ms = 0;
            p_transfer_data->fp_transfer_aborted_callback = CLD_NULL;
            p_transfer_data->callback.fp_usb_in_transfer_complete = CLD_NULL;
            rv = CLD_USB_TRANSFER_ACCEPT;
        }
        else if (p_req_params->req == CLD_REQ_RANGE)
        {
            p_transfer_data->p_data_buffer = (unsigned char *)&user_audio_get_clock_sample_rate_range_resp;
            p_transfer_data->num_bytes = sizeof(user_audio_get_clock_sample_rate_range_resp);
            p_transfer_data->transfer_timeout_ms = 0;
            p_transfer_data->fp_transfer_aborted_callback = CLD_NULL;
            p_transfer_data->callback.fp_usb_in_transfer_complete = CLD_NULL;
            rv = CLD_USB_TRANSFER_ACCEPT;
        }
    }
    else if (p_req_params->entity_id == USER_AUDIO_CLOCK_SOURCE_ID_ADC)
    {
        if (p_req_params->req == CLD_REQ_CURRENT)
        {
            p_transfer_data->p_data_buffer = (unsigned char *)&user_audio_data.clock_sample_rate;
            p_transfer_data->num_bytes = 4;
            p_transfer_data->transfer_timeout_ms = 0;
            p_transfer_data->fp_transfer_aborted_callback = CLD_NULL;
            p_transfer_data->callback.fp_usb_in_transfer_complete = CLD_NULL;
            rv = CLD_USB_TRANSFER_ACCEPT;
        }
        else if (p_req_params->req == CLD_REQ_RANGE)
        {
            p_transfer_data->p_data_buffer = (unsigned char *)&user_audio_get_clock_sample_rate_range_resp;
            p_transfer_data->num_bytes = sizeof(user_audio_get_clock_sample_rate_range_resp);
            p_transfer_data->transfer_timeout_ms = 0;
            p_transfer_data->fp_transfer_aborted_callback = CLD_NULL;
            p_transfer_data->callback.fp_usb_in_transfer_complete = CLD_NULL;
            rv = CLD_USB_TRANSFER_ACCEPT;
        }
    }
    return rv;
}

/**
 * Function called when the Isochronous OUT interface is enabled/disabled by the
 * USB Host using the Set Interface Alternate Setting request.
 *
 * @param enabled - CLD_TRUE = Isochronous OUT endpoint Enabled.
 */
static void user_audio_streaming_rx_endpoint_enabled (CLD_Boolean enabled)
{
    user_audio_data.isochronous_out_enabled = enabled;

    if (enabled == CLD_TRUE)
    {
        first_feedback = CLD_TRUE;
        user_audio_stream_data_receive(); /* Prime the OUT endpoint to receive the first steaming audio data packet. */
    }
    else
    {
        user_audio_data.rate_feedback_idle = CLD_TRUE;
        ll_mgr_reset(&user_audio_out_ll);
        if (p_user_audio_out_ll_node)
        {
            ll_mgr_free_node(p_user_audio_out_ll_node);
        }
    }
}

/**
 * Function called when the Isochronous IN interface is enabled/disabled by the
 * USB Host using the Set Interface Alternate Setting request.
 *
 * @param enabled - CLD_TRUE = Isochronous IN endpoint Enabled.
 */
static void user_audio_streaming_tx_endpoint_enabled (CLD_Boolean enabled)
{
    unsigned int i;
    user_audio_data.isochronous_in_enabled = enabled;

    if (enabled == CLD_TRUE)
    {
        user_audio_num_in_packets = 0;

        /* If there isn't any data waiting to be sent to the Host send a zero
           packet. */
        if (user_audio_in_ll.num_elements == 0)
        {
            p_user_audio_zero_ll_node = ll_mgr_get_node(0);

            if (p_user_audio_zero_ll_node)
            {
                p_user_audio_zero_ll_node->size = AUDIO_IN_BUFFER_SIZE;

                for (i = 0; i < AUDIO_IN_BUFFER_SIZE; i++)
                {
                    p_user_audio_zero_ll_node->data[i] = 0;
                }

                user_audio_transmit_audio_stream_data(p_user_audio_zero_ll_node);
            }
        }
    }
    else
    {
        ll_mgr_reset(&user_audio_in_ll);
        if (p_user_audio_in_ll_node)
        {
            ll_mgr_free_node(p_user_audio_in_ll_node);
        }
        user_audio_num_in_packets = 0;

        user_audio_data.isochronous_in_idle = CLD_TRUE;
    }
}

/**
 * Adds the data buffer passed to the function to the head of the microphone circular queue.
 *
 * @param p_data - Pointer to the data being added to the microphone data queue
 * @param data_length - microphone data length.
 *
 */
void user_audio_transmit_audio_stream_data (LL_Mgr_Node * p_node)
{
    if (user_audio_data.isochronous_in_enabled == CLD_TRUE)
    {
        /* only load four packets before one is sent to the USB Host after being enabled.
           This is to avoid overflowing the linked list between the time the Audio IN is enabled
           and the Host starts sending IN tokens. */
        if (((user_audio_num_in_packets < 4) || (user_audio_num_in_packets == 0xff)) && (user_audio_in_ll.num_elements <= 6))
        {
            ll_mgr_insert(&user_audio_in_ll, p_node);
            user_audio_num_in_packets++;

            if (user_audio_data.isochronous_in_idle == CLD_TRUE)
            {
                user_audio_tx_audio_data();
            }
        }
        else
        {
            ll_mgr_free_node(p_node);
        }
    }
    else
    {
        ll_mgr_free_node(p_node);
    }
}

/**
 * This function returns the number elements in the audio IN linked list.
 */
unsigned char user_audio_audio_in_get_num_queued_packets (void)
{
    return user_audio_in_ll.num_elements;
}

/**
 * This function is called by the CLD library when Bulk IN serial data has been
 * transmitted to the Host.
 */
static void user_cdc_usb_in_transfer_complete (void)
{
}

/**
 * This function is called by the CLD library when Bulk IN serial data has been
 * aborted.
 */
static void user_cdc_usb_in_transfer_aborted (void)
{
}

/**
 * Function called when Serial data has been received.
 *
 * @retval CLD_USB_DATA_GOOD
 * @retval CLD_USB_DATA_BAD_STALL
 */
static CLD_USB_Data_Received_Return_Type user_cdc_serial_data_rx_transfer_complete (unsigned int num_bytes)
{
    /* Transmit the received data using the SC59x UART. */
    user_cdc_tx_serial_data(num_bytes+sizeof(user_cdc_serial_data_rx_buffer.resp_hdr), (unsigned char*)&user_cdc_serial_data_rx_buffer);
    user_cdc_serial_data_receive();

    return CLD_USB_DATA_GOOD;
}

/**
 * This function is called by the CLD library when Bulk OUT
 * serial data has been aborted.
 */
static void user_cdc_usb_out_transfer_aborted (void)
{
    user_cdc_serial_data_receive();
}

/**
 * This function primes the Bulk OUT endpoint so it will
 * receive data.
 */
static void user_cdc_serial_data_receive (void)
{
    static CLD_USB_Transfer_Params transfer_params;
    CLD_USB_Data_Receive_Return_Type rv;

    transfer_params.num_bytes = USER_CDC_SERIAL_BUFFER_SIZE;
    transfer_params.p_data_buffer = user_cdc_serial_data_rx_buffer.data;
    transfer_params.transfer_timeout_ms = 0;
    transfer_params.fp_transfer_aborted_callback = user_cdc_usb_out_transfer_aborted;
    transfer_params.callback.fp_usb_out_transfer_complete = user_cdc_serial_data_rx_transfer_complete;

    rv = cld_cdc_lib_receive_serial_data(&transfer_params);
    if (rv == CLD_USB_RECEIVE_FAILED_MISALIGNED)
    {
        /* CLD library requires transfer_params.p_data_buffer to be 32-bit aligned */
    }
    else if (rv == CLD_USB_RECEIVE_FAILED)
    {
        /* Bulk OUT endpoint busy */
    }
}

/**
 * This function is called by the CLD library when a CDC Send Encapsulated Command
 * request is received. This function sets the p_transfer_data parameters to
 * select where the received data should be stored, and what function should
 * be called when the transfer is complete.
 *
 * @param p_transfer_data                - Pointer to the Control OUT transfer data.
 * @param p_transfer_data->num_bytes     - Number of Control OUT bytes.
 * @param p_transfer_data->p_data_buffer - Set to the address where to
 *                                         store the Control OUT data.
 * @param p_transfer_data->callback.usb_in_transfer_complete -
 *                                         Function called when the
 *                                         Control OUT bytes have been
 *                                         received.
 * @param p_transfer_data->transfer_aborted_callback -
 *                                         Optional function that is
 *                                         called if the transfer is
 *                                         aborted.
 * @retval CLD_USB_TRANSFER_ACCEPT - Store the data using the p_transfer_data
 *                                  parameters.
 * @retval CLD_USB_TRANSFER_PAUSE - The device isn't ready to process this
 *                                  out packet so pause the transfer
 *                                  until the cld_sc57x_audio_2_0_w_cdc_lib_resume_paused_control_transfer
 *                                  function is called.
 * @retval CLD_USB_TRANSFER_DISCARD - Discard this packet.
 * @retval CLD_USB_TRANSFER_STALL - Stall the OUT endpoint.
 */
static CLD_USB_Transfer_Request_Return_Type user_cdc_cmd_send_encapsulated_cmd (CLD_USB_Transfer_Params * p_transfer_data)
{
    /* Not currently supported by this example. */
    return CLD_USB_TRANSFER_STALL;
}

/**
 * This function is called by the CLD library when a CDC Get Encapsulated Command
 * request is received. This function sets the p_transfer_data parameters to
 * select where the transmit data should be sourced, and what function should
 * be called when the transfer is complete.
 *
 * @param p_transfer_data                - Pointer to the Control IN transfer data.
 * @param p_transfer_data->num_bytes     - Number of Control IN bytes.
 * @param p_transfer_data->p_data_buffer - Set to the address where to
 *                                         source the Control IN data.
 * @param p_transfer_data->callback.usb_in_transfer_complete -
 *                                         Function called when the
 *                                         Control IN bytes have been
 *                                         sent.
 * @param p_transfer_data->transfer_aborted_callback -
 *                                         Optional function that is
 *                                         called if the transfer is
 *                                         aborted.
 * @retval CLD_USB_TRANSFER_ACCEPT - Store the data using the p_transfer_data
 *                                   parameters.
 * @retval CLD_USB_TRANSFER_PAUSE - The device isn't ready to process this
 *                                  out packet so pause the transfer
 *                                  until the cld_sc57x_audio_2_0_w_cdc_lib_resume_paused_control_transfer
 *                                  function is called.
 * @retval CLD_USB_TRANSFER_DISCARD - Discard this packet.
 * @retval CLD_USB_TRANSFER_STALL - Stall the OUT endpoint.
 */
static CLD_USB_Transfer_Request_Return_Type user_cdc_cmd_get_encapsulated_resp (CLD_USB_Transfer_Params * p_transfer_data)
{
    /* Not currently supported by this example. */
    return CLD_USB_TRANSFER_STALL;
}

/**
 * This function is called by the CLD library when a CDC Set Line Coding
 * request is received. This function saves the requested line coding.
 *
 * @retval CLD_USB_DATA_GOOD
 * @retval CLD_USB_DATA_BAD_STALL
 */
static CLD_USB_Data_Received_Return_Type user_cdc_cmd_set_line_coding (CLD_CDC_Line_Coding * p_line_coding)
{
    user_cdc_line_coding = *p_line_coding;
    return CLD_USB_DATA_GOOD;
}

/**
 * This function is called by the CLD library when a CDC Get Line Coding
 * request is received. This function loads the current line coding into the
 * p_line_coding structure passed to the function.
 *
 * @retval CLD_SUCCESS - There was no errors copying the current line coding.
 * @retval CLD_FAIL    - There was a problem copying the current line coding.
 */
static CLD_RV user_cdc_cmd_get_line_coding (CLD_CDC_Line_Coding * p_line_coding)
{
    *p_line_coding = user_cdc_line_coding;

    return CLD_SUCCESS;
}

/**
 * This function is called by the CLD library when a CDC Set Control Line State
 * request is received.
 *
 * @retval CLD_USB_DATA_GOOD - Set control line state parameters accepted.
 * @retval CLD_USB_DATA_BAD_STALL - An error occurred while applying the requested
 *                                  set control line state parameters.
 */
static CLD_USB_Data_Received_Return_Type user_cdc_cmd_set_control_line_state (CLD_CDC_Control_Line_State * p_control_line_state)
{
    return CLD_USB_DATA_GOOD;
}

/**
 * This function is called by the CLD library when a CDC Send Break
 * request is received.
 *
 * @retval CLD_USB_DATA_GOOD
 * @retval CLD_USB_DATA_BAD_STALL
 */
static CLD_USB_Data_Received_Return_Type user_cdc_cmd_send_break (unsigned short duration)
{
    return CLD_USB_DATA_GOOD;
}

/**
 * Transmits the specified data to the CDC Host using the serial data Bulk IN endpoint.
 *
 * @param length   - number of bytes to transmit.
 * @param p_buffer - pointer to the data buffer to transmit
 */
static CLD_USB_Data_Transmit_Return_Type user_cdc_tx_serial_data (unsigned long length, unsigned char * p_buffer)
{
    static CLD_USB_Transfer_Params transfer_params;
    transfer_params.num_bytes = length;
    transfer_params.p_data_buffer = p_buffer;
    transfer_params.callback.fp_usb_in_transfer_complete = user_cdc_usb_in_transfer_complete;
    transfer_params.fp_transfer_aborted_callback = user_cdc_usb_in_transfer_aborted;
    transfer_params.transfer_timeout_ms = 1000;

    return cld_cdc_lib_transmit_serial_data (&transfer_params);
}

/**
 * Function called when a USB event occurs on the USB Port with USB Audio v2.0 and CDC.
 *
 * @param event - identifies which USB event has occurred.
 */
static void user_usb_event (CLD_USB_Event event)
{
    switch (event)
    {
        case CLD_USB_CABLE_CONNECTED:
            adi_gpio_Set(ADI_GPIO_PORT_C, ADI_GPIO_PIN_1);
        break;
        case CLD_USB_CABLE_DISCONNECTED:
            adi_gpio_Clear(ADI_GPIO_PORT_C, ADI_GPIO_PIN_1);
            user_audio_streaming_tx_endpoint_enabled(CLD_FALSE);
            user_audio_streaming_rx_endpoint_enabled(CLD_FALSE);
        break;
        case CLD_USB_ENUMERATED_CONFIGURED_HS:
        case CLD_USB_ENUMERATED_CONFIGURED_FS:
            user_cdc_serial_data_receive();
        break;
        case CLD_USB_UN_CONFIGURED:
            user_audio_streaming_tx_endpoint_enabled(CLD_FALSE);
            user_audio_streaming_rx_endpoint_enabled(CLD_FALSE);
        break;
        case CLD_USB_BUS_RESET:
            user_audio_streaming_tx_endpoint_enabled(CLD_FALSE);
            user_audio_streaming_rx_endpoint_enabled(CLD_FALSE);
        break;
    }
}

/**
 * Function called when the usb feedback endpoint data has been transmitted.
 *
 * @param event - identifies which USB event has occurred.
 */
static void user_audio_feedback_xfr_done (void)
{
    user_audio_data.rate_feedback_idle = CLD_TRUE;
    first_feedback = CLD_FALSE;

    user_audio_tx_feedback_data();
}

/**
 * Returns the oldest element in the USB OUT linked list.
 *
 * @retval Pointer to the oldest element in the USB OUT linked list
 */
LL_Mgr_Node * user_audio_w_cdc_get_audio_out_data (void)
{
    LL_Mgr_Node * rv = ll_mgr_remove(&user_audio_out_ll);

    return rv;
}

/**
 * Returns the number of elements in the USB OUT linked list.
 *
 * @retval Number of elements in the USB OUT linked list
 */
unsigned long user_audio_w_cdc_get_audio_out_data_num_elements (void)
{
    return user_audio_out_ll.num_elements;
}

/**
 * User defined USB 0 ISR.
 *
 * @param Event, pArg.
 */
static void user_usb_lib_usb_isr(uint32_t Event, void *pArg)
{
    cld_usb0_isr_callback();
}

/**
 * User defined Timer ISR.
 *
 * @param pCBParam, Event, pArg.
 */
void user_usb_lib_timer_handler(void *pCBParam, uint32_t Event, void *pArg)
{
    switch(Event)
    {
        case ADI_TMR_EVENT_DATA_INT:
            /* Call the CLD library 125 microsecond function */
            cld_time_125us_tick();
            break;

        default:
            break;
    }

    return;
}
/**
 * User defined Timer ISR - currently set to run every millisecond.
 *
 * @param pCBParam, Event, pArg.
 */
void user_usb_lib_low_priority_timer_handler(void *pCBParam, uint32_t Event, void *pArg)
{
    switch(Event)
    {
        case ADI_TMR_EVENT_DATA_INT:

            break;

        default:
            break;
    }

    return;
}

/**
 * Function called when the CLD library reports a status.
 *
 * @param status_code, p_additional_data, additional_data_size.
 */
static void user_usb_lib_status (unsigned short status_code, void * p_additional_data, unsigned short additional_data_size)
{
    /* Process the CLD library status */
    //char * p_str = cld_lib_status_decode(status_code, p_additional_data, additional_data_size);
}

/**
 * Function called when the system services uart receives a byte.
 *
 * @param byte.
 */
void user_usb_lib_rx_term_byte (unsigned char byte)
{
    switch (byte)
    {

    }
}

/**.
 *  Function used to initialize and reset the USB Phy.
 * @return CLD_ONGOING results in this function getting additional runtime.
 * @return CLD_SUCCESS USB Phy initialized successfully.
 * @return CLD_FAIL Phy initialization failed, USB library initialization failure
 */
static CLD_RV user_init_usb_phy (void)
{
    static unsigned char phy_init_state = 0;
    static CLD_Time phy_init_time;
    CLD_RV rv = CLD_ONGOING;

    switch (phy_init_state)
    {
        case 0:
            /* Configure the PHY reset PIN */
            adi_gpio_SetDirection(ADI_GPIO_PORT_G, ADI_GPIO_PIN_11, ADI_GPIO_DIRECTION_OUTPUT);
            /* Assert the reset PIN */
            adi_gpio_Clear(ADI_GPIO_PORT_G, ADI_GPIO_PIN_11);

            phy_init_time = cld_time_get();
            phy_init_state++;
        break;
        case 1:
            if (cld_time_passed_ms(phy_init_time) > 10)
            {
                /* Unassert reset */
                adi_gpio_Set(ADI_GPIO_PORT_G, ADI_GPIO_PIN_11);
                phy_init_time = cld_time_get();
                phy_init_state++;
            }
        break;
        case 2:
            /* After releasing reset the USB3340 has a Tstart time of 1.2ms Max */
            if (cld_time_passed_ms(phy_init_time) > 10)
            {
                rv = CLD_SUCCESS;
            }
        break;
    }
    return rv;
}

